# -*- coding: utf-8 -*-
"""

Partially taken from:
    https://github.com/mitsuhiko/flask/blob/master/flask/_compat.py

    flask._compat
    ~~~~~~~~~~~~~
    Some py2/py3 compatibility support based on a stripped down
    version of six so we don't have to depend on a specific version
    of it.
    :copyright: (c) 2015 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.

"""
import sys

PY2 = sys.version_info[0] == 2

if not PY2:
    string_types = (str,)

    from io import StringIO

else:
    string_types = (str, unicode)

    from cStringIO import StringIO



from collections import namedtuple
from enum import Enum
import re


_COMMENT_TEST_RX = re.compile(r'^\s*#.*$')
_CTX_WS_RX = re.compile(r'\S')
_INT_RX = re.compile(r'-?\d+')


class Keywords(object):
    css_start_delim = '$'
    accessor_start_delim = '|'
    attr_accessor_start = '['
    attr_accessor_end = ']'
    text_accessor = 'text'


TokenType = Enum('TokenType', ' '.join((
    'Context',
    'QueryStatement',
    'QueryStatementEnd',
    'CSSSelector',
    'AccessorSequence',
    'IndexAccessor',
    'TextAccessor',
    'AttrAccessor',
    'DirectiveStatement',
    'DirectiveIdentifier',
    'DirectiveBodyItem',
    )))


class Token(namedtuple('Token', 'type_ content line line_num start end')):
    __slots__ = ()

    def __repr__(self):
        return ('Token: {{\n'
                '        type_: {!r}\n'
                '      content: "{}"\n'
                '     line num: {}\n'
                '   start, end: {}, {}\n'
                '         line: "{}"\n'
                '}}').format(self.type_, self.content, self.line_num, self.start, self.end, self.line)


class ScanError(Exception):

    @staticmethod
    def make(scanner, msg, extra=None):
        return ScanError(msg, scanner.line, scanner.line_num, scanner.pos, extra)

    def __init__(self, message, line, line_num, pos, extra=None):
        self.message = message
        self.line = line
        self.line_num = line_num
        self.pos = pos
        self.extra = extra

    def __str__(self):
        return ('ScanError: {{\n'
                '      message: {!r}\n'
                '         line: "{}"\n'
                '                {}^\n'
                '     line num: {}\n'
                '          pos: {}\n'
                '        extra: {!r}\n'
                '}}').format(self.message, self.line, ' ' * self.pos,
                             self.line_num, self.pos, self.extra)


class Scanner(object):

    def __init__(self, fobj):
        self.fobj = fobj
        self.line = None
        self.line_num = 0
        self.start = 0
        self.pos = 0

    @property
    def _c(self):
        return self.line[self.pos]

    @property
    def _eol(self):
        return self.pos >= len(self.line)

    @property
    def _to_eol_content(self):
        return self.line[self.start:]

    @property
    def _tok_content(self):
        return self.line[self.start:self.pos]

    def scan(self):
        for line in self.fobj:
            self.line_num += 1
            self.start = self.pos = 0
            self.line = line.rstrip()
            if not self.line or _COMMENT_TEST_RX.match(self.line):
                continue
            scan_fn = self._scan_context
            while scan_fn:
                scan_fn, tok = scan_fn()
                if tok:
                    yield tok

    def _ignore(self):
        self.start = self.pos

    def _make_token(self, type_):
        tok = Token(type_,
                    self.line[self.start:self.pos],
                    self.line,
                    self.line_num,
                    self.start,
                    self.pos)
        self.start = self.pos
        return tok

    def _make_marker_token(self, type_):
        """Make a token that has no content"""
        tok = Token(type_,
                    '',
                    self.line,
                    self.line_num,
                    self.start,
                    self.start)
        return tok


    def _accept(self, valid=None, alpha=False, num=False, alphanum=False):
        if self._eol:
            return False
        ok = False
        if valid and self._c in valid:
            ok = True
        elif alpha and self._c.isalpha():
            ok = True
        elif num and self._c.isdigit():
            ok = True
        elif alphanum and self._c.isalnum():
            ok = True
        elif not valid and not alpha and not num and not alphanum:
            raise ScanError.make(self, 'Invalid _accept call, valid or preset not specified')
        if ok:
            self.pos += 1
        return ok

    def _accept_run(self, valid=None, alpha=False, num=False, alphanum=False):
        count = 0
        tests = []
        if valid:
            tests.append(lambda: self._c in valid)
        if alpha:
            tests.append(lambda: self._c.isalpha())
        if num:
            tests.append(lambda: self._c.isdigit())
        if alphanum:
            tests.append(lambda: self._c.isalnum())
        if not tests:
            raise ScanError.make(self, 'Invalid _accept call, valid or preset not specified')
        while not self._eol and any(t() for t in tests):
            count += 1
            self.pos += 1
        return count

    def _accept_until(self, one_of):
        count = 0
        while not self._eol and self._c not in one_of:
            count += 1
            self.pos += 1
        return count

    def _consume(self, val):
        if self._eol:
            return False
        next_pos = self.pos + len(val)
        if self.line[self.pos:next_pos] == val:
            self.pos = next_pos
            return True
        return False

    def _consume_re(self, test_re):
        if self._eol:
            return False
        m = test_re.match(self._to_eol_content)
        if not m:
            return False
        self.pos += m.end()
        return True


    def _scan_context(self):
        m = _CTX_WS_RX.search(self.line)
        if not m:
            raise ScanError.make(self, 'Invalid state: error parsing context')
        self.pos = m.start()
        tok = self._make_token(TokenType.Context)
        return self._scan_statement, tok

    def _scan_statement(self):
        if self._c.isalpha():
            tok = self._make_marker_token(TokenType.DirectiveStatement)
            return self._scan_directive, tok
        else:
            tok = self._make_marker_token(TokenType.QueryStatement)
            return self._scan_query_statement, tok

    def _scan_directive(self):
        ok = self._accept(alpha=True)
        if not ok:
            raise ScanError.make(self, 'Invalid directive, must start with an alpha, not %s' % self._c)
        if self._accept_until(':') < 1:
            raise ScanError.make(self, 'Invalid directive, 0 length')
        tok = self._make_token(TokenType.DirectiveIdentifier)
        if not self._accept(':'):
            raise ScanError.make(self, 'Invalid directive identifier terminator: "%s", ":" required.' % self._tok_content)
        self._ignore()
        return self._scan_directive_body, tok

    def _scan_directive_body(self):
        self._accept_run(' ')
        self._ignore()
        if self._accept_until(' ') < 1:
            raise ScanError.make(self, 'Invalid directive body item, 0 length')
        tok = self._make_token(TokenType.DirectiveBodyItem)
        return None if self._eol else self._scan_directive_body, tok

    def _scan_query_statement(self):
        if self._c == Keywords.css_start_delim:
            return self._scan_css_selector()
        elif self._c == Keywords.accessor_start_delim:
            return self._scan_accessor_sequence()
        else:
            raise ScanError.make(self, 'Invalid query statement: %s' % self._to_eol_content)

    def _end_query_statement(self):
        tok = self._make_marker_token(TokenType.QueryStatementEnd)
        return None, tok

    def _scan_css_selector(self):
        self._accept(Keywords.css_start_delim)
        self._accept_run(' ')
        self._ignore()
        if self._accept_until(Keywords.accessor_start_delim) < 1:
            raise ScanError.make(self, 'Invalid CSS Selector: %s' % self._to_eol_content)
        tok = self._make_token(TokenType.CSSSelector)
        if self._eol:
            return self._end_query_statement, tok
        elif self._c == Keywords.accessor_start_delim:
            return self._scan_accessor_sequence, tok
        else:
            raise ScanError.make(self, 'EOL or accessor sequence expected, instead found %s' % self._to_eol_content)

    def _scan_accessor_sequence(self):
        # create the marker token at the start of the sequence
        tok = self._make_marker_token(TokenType.AccessorSequence)
        # skip the initial marker character
        self._accept(Keywords.accessor_start_delim)
        self._ignore()
        return self._scan_accessor, tok

    def _scan_accessor(self):
        self._accept_run(' ')
        self._ignore()
        if self._eol:
            return self._end_query_statement()
        elif self._c == Keywords.attr_accessor_start:
            # attribute accessor
            self._accept_until(Keywords.attr_accessor_end)
            if not self._accept(Keywords.attr_accessor_end):
                raise ScanError.make(self, 'Invalid Attr Accessor: %s' % self._to_eol_content)
            tok = self._make_token(TokenType.AttrAccessor)
            return self._scan_accessor, tok
        elif self._consume(Keywords.text_accessor):
            # text accessor
            tok = self._make_token(TokenType.TextAccessor)
            return self._scan_accessor, tok
        else:
            # index accesor, test for a valid number, ie -?\d+
            if not self._consume_re(_INT_RX):
                raise ScanError.make(self, 'Expected an accessor, instead found: %s' % self._to_eol_content)
            tok = self._make_token(TokenType.IndexAccessor)
            return self._scan_accessor, tok



from pyquery import PyQuery

from .parser import parse


class TakeTemplate(object):

    @staticmethod
    def from_file(path):
        with open(path, 'rb') as f:
            return TakeTemplate(f.read().decode('utf-8'))

    def __init__(self, src, **kwargs):
        self.node = parse(src)
        self.base_url = kwargs.get('base_url', None)

    def take(self, *args, **kwargs):
        base_url = kwargs.pop('base_url', None) or self.base_url
        _doc = PyQuery(*args, **kwargs)
        if base_url:
            _doc.make_links_absolute(base_url)
        rv = {}
        self.node.do(None, rv=rv, value=_doc, last_value=_doc)
        return rv

    def __call__(self, *args, **kwargs):
        return self.take(*args, **kwargs)



from collections import namedtuple

from pyquery import PyQuery

from ._compat import string_types, StringIO
from .scanner import Scanner, TokenType


class AlreadyParsedError(Exception):
    pass

class UnexpectedEOFError(Exception):
    pass

class UnexpectedTokenError(Exception):
    def __init__(self, found, expected, msg=None, token=None):
        self.found = found
        self.expected = expected
        self.msg = msg
        self.token = token

    def __str__(self):
        return ('UnexpectedTokenError {{\n'
                '       found: {!r}\n'
                '    expected: {!r}\n'
                '         msg: {!r}\n'
                '       token: {}\n'
                '}}').format(self.found, self.expected, self.msg, self.token)

class InvalidDirectiveError(Exception):
    def __init__(self, ident, msg=None):
        self.ident = ident
        self.msg = msg

    def __str__(self):
        return ('InvalidDirectiveError {{\n'
                '      ident: {!r}\n'
                '    message: {!r}\n'
                '}}').format(self.ident, self.message)

class TakeSyntaxError(Exception):
    def __init__(self, msg, extra=None):
        self.msg = msg
        self.extra = extra

    def __str__(self):
        return ('TakeSyntaxError {{\n'
                '      message: {!r}\n'
                '        extra: {!r}\n'
                '}}').format(self.message, self.extra)


def ensure_pq(elm):
    if isinstance(elm, PyQuery):
        return elm
    else:
        return PyQuery(elm)

def make_css_query(selector):
    return lambda elm: ensure_pq(elm)(selector)

def make_index_query(index_str):
    index = int(index_str)
    if index > -1:
        return lambda elm: ensure_pq(elm).eq(index)
    # PyQuery doesn't handle negative indexes, so calc the real index each time
    def index_query(elm):
        elm = ensure_pq(elm)
        i = len(elm) + index
        return elm.eq(i)
    return index_query

def make_text_query():
    return lambda elm: ensure_pq(elm).text()

def make_attr_query(attr):
    return lambda elm: ensure_pq(elm).attr(attr)

def parse_save_to_id(tok):
    if tok.type_ != TokenType.DirectiveBodyItem:
        raise UnexpectedTokenError(tok.type_, TokenType.DirectiveBodyItem)
    save_to = tok.content.strip()
    if '.' in save_to:
        return tuple(save_to.split('.'))
    else:
        return (save_to,)

def save_to(dest, name_parts, value):
    """
    Util to save some name sequence to a dict. For instance, `("location","query")` would save
    to dest["location"]["query"].
    """
    if len(name_parts) > 1:
        for part in name_parts[:-1]:
            if part not in dest:
                dest[part] = {}
            dest = dest[part]
    dest[name_parts[-1]] = value


class ContextNode(object):
    __slots__ = ('__depth', '__nodes', '__rv', '__value', 'last_value')

    def __init__(self, depth, nodes):
        self.__depth = depth
        self.__nodes = nodes
        self.__rv = None
        self.__value = None
        self.last_value = None

    @property
    def rv(self):
        return self.__rv

    @property
    def value(self):
        return self.__value

    def do(self, context, rv=None, value=None, last_value=None):
        self.__rv = rv if rv != None else context.rv
        # value in a sub-context is derived from the last_value in the parent context
        self.__value = value if value != None else context.last_value
        if last_value != None:
            self.last_value = last_value
        elif value != None:
            self.last_value = value
        elif context:
            if context.last_value != None:
                self.last_value = context.last_value
            elif context.value != None:
                self.last_value = context.value
            else:
                self.last_value = None
        else:
            self.last_value = None
        for node in self.__nodes:
            node.do(self)


class QueryNode(namedtuple('QueryNode', 'queries')):
    __slots__ = ()
    def do(self, context):
        # start last_value based on context.value
        val = context.value
        # queries can be a sequence of queries
        for query in self.queries:
            val = query(val)
        # update context.last_value
        context.last_value = val


class SaveNode(namedtuple('SaveNode', 'ident_parts')):
    __slots__ = ()
    def do(self, context):
        save_to(context.rv, self.ident_parts, context.value)


class SaveEachNode(namedtuple('SaveEachNode', 'ident_parts sub_ctx_node')):
    __slots__ = ()
    def do(self, context):
        results = []
        save_to(context.rv, self.ident_parts, results)
        for item in context.value:
            rv = {}
            results.append(rv)
            self.sub_ctx_node.do(None, rv, item, item)


class ContextParser(object):

    def __init__(self, depth, tok_gen):
        self._depth = depth
        self._tok_generator = tok_gen
        self._nodes = None
        self._tok = None
        self._is_done = False

    def destroy(self):
        self._depth = None
        self._tok_generator = None
        self._nodes = None
        self._tok = None
        self._is_done = None


    def parse(self):
        if self._is_done:
            raise AlreadyParsedError
        self._nodes = []
        tok = self._parse()
        return ContextNode(self._depth, self._nodes), tok


    def _next_tok(self, eof_errors=True):
        try:
            self._tok = next(self._tok_generator)
            return self._tok
        except StopIteration:
            self._tok = None
            self._is_done = True
            if eof_errors:
                raise UnexpectedEOFError

    def _parse(self):
        while True:
            tok = self._next_tok()
            if tok.type_ == TokenType.QueryStatement:
                self._parse_query()
                tok = None
            elif tok.type_ == TokenType.DirectiveStatement:
                # some directives consume a context token to determine if they end (eg SaveEachNode)
                tok = self._parse_directive()
            else:
                raise UnexpectedTokenError(tok.type_,
                                          (TokenType.QueryStatement, TokenType.DirectiveStatement),
                                          token=tok)
            if not tok:
                tok = self._next_tok(eof_errors=False)
            if self._is_done:
                # returning None means EOF ended this context, not a context exit
                return
            if tok.type_ != TokenType.Context:
                raise UnexpectedTokenError(self._tok.type_, TokenType.Context, token=tok)
            if tok.end > self._depth:
                end_tok = self._parse_context()
                if not end_tok:
                    # end_tok is either the last token grabbed by the sub-context or None, if it is
                    # None, then we have reached EOF
                    return
                else:
                    # if end_tok a token, it is a context token and is less than the sub-context,
                    # possibly less the current context... use it as the context node to check to
                    # see if this context should end
                    tok = end_tok
                # TODO(joe): consider checking for bug/bad contexts in the source where enter the
                # sub-context with indent+4 but exit with indent+2, so, exit the sub-context but
                # would be in another sub-context (prob should not allow)
            if tok.end < self._depth:
                return tok
            # context token same depth as this context, so continue parsing

    def _parse_context(self):
        sub_ctx = ContextParser(self._tok.end, self._tok_generator)
        sub_ctx_node, tok = sub_ctx.parse()
        self._nodes.append(sub_ctx_node)
        sub_ctx.destroy()
        return tok

    def _parse_query(self):
        self._next_tok()
        if self._tok.type_ == TokenType.CSSSelector:
            queries = self._parse_css_selector()
        elif self._tok.type_ == TokenType.AccessorSequence:
            queries = self._parse_accessor_seq()
        else:
            raise UnexpectedTokenError(self._tok.type_, (TokenType.CSSSelector,
                                                         TokenType.AccessorSequence))
        node = QueryNode(queries)
        self._nodes.append(node)

    def _parse_css_selector(self):
        selector = self._tok.content.strip()
        # expects a valid css selector
        query = make_css_query(selector)
        self._next_tok()
        if self._tok.type_ == TokenType.QueryStatementEnd:
            return (query,)
        elif self._tok.type_ == TokenType.AccessorSequence:
            return (query,) + self._parse_accessor_seq()
        else:
            raise UnexpectedTokenError(self._tok.type_, (TokenType.QueryStatementEnd,
                                                         TokenType.AccessorSequence))

    def _parse_accessor_seq(self):
        queries = ()
        tok = self._next_tok()
        # index accessor has to be first
        if tok.type_ == TokenType.IndexAccessor:
            index_str = tok.content
            queries = (make_index_query(index_str),)
            tok = self._next_tok()
            # index accessor can be the only accessor
            if tok.type_ == TokenType.QueryStatementEnd:
                return queries
        if tok.type_ == TokenType.TextAccessor:
            queries += (make_text_query(),)
        elif tok.type_ == TokenType.AttrAccessor:
            # strip spaces and remove the brackets
            # ex: [src]
            attr = tok.content.strip()[1:-1]
            queries += (make_attr_query(attr),)
        else:
            # if it got here, something is wrong, should either have exited after the index
            # accessor (if there was one) or there should have been a text or attr accessor
            expected = (TokenType.TextAccessor, TokenType.AttrAccessor)
            if not queries:
                expected += (TokenType.IndexAccessor,)
            raise UnexpectedTokenError(tok.type_, expected, token=tok)
        # can only have one of text accessor or attr accessor, so do not need to loop
        tok = self._next_tok()
        if tok.type_ != TokenType.QueryStatementEnd:
            raise UnexpectedTokenError(tok.type_, TokenType.QueryStatementEnd, token=tok)
        return queries

    def _parse_directive(self):
        tok = self._next_tok()
        if tok.type_ != TokenType.DirectiveIdentifier:
            raise UnexpectedTokenError(tok.type_, TokenType.DirectiveIdentifier, token=tok)
        dir_ident = tok.content.strip()
        if dir_ident == 'save':
            return self._parse_save_directive()
        elif dir_ident == 'save each':
            return self._parse_save_each_directive()
        else:
            raise InvalidDirectiveError(dir_ident, 'Unknown directive, valid directives: %r' %
                                                   ['save', 'save each'])

    def _parse_save_directive(self):
        tok = self._next_tok()
        save_id_parts = parse_save_to_id(tok)
        node = SaveNode(save_id_parts)
        self._nodes.append(node)

    def _parse_save_each_directive(self):
        tok = self._next_tok()
        save_id_parts = parse_save_to_id(tok)
        # consume the context token and parse the sub-context SaveEachNode will manage
        tok = self._next_tok()
        if tok.type_ != TokenType.Context:
            raise UnexpectedTokenError(tok.type_, TokenType.Context, token=tok)
        if tok.end <= self._depth:
            raise TakeSyntaxError('Invalid depth, expecting to start a "save each" context.',
                                  extra=tok)
        sub_ctx = ContextParser(tok.end, self._tok_generator)
        sub_ctx_node, tok = sub_ctx.parse()
        sub_ctx.destroy()
        node = SaveEachNode(save_id_parts, sub_ctx_node)
        self._nodes.append(node)
        return tok


def parse(src):
    if isinstance(src, string_types):
        fobj = StringIO(src)
    else:
        fobj = src
    scanner = Scanner(fobj)
    tok_generator = scanner.scan()
    tok = next(tok_generator)
    if tok.type_ != TokenType.Context:
        raise UnexpectedTokenError(tok.type_, TokenType.Context, 'Leading context token not found')
    ctx_parser = ContextParser(tok.end, tok_generator)
    node, last_tok = ctx_parser.parse()
    ctx_parser.destroy()
    if last_tok:
        raise UnexpectedTokenError(last_tok, 'EOF')
    return node



# -*- coding: utf-8 -*-
"""
A DSL for extracting data from a web page. The DSL serves two purposes:
finds elements and extracts their text or attribute values. The main
reason for developing this is to have all the CSS selectors for scraping
a site in one place (I prefer CSS selectors over anything else).

The DSL wraps `PyQuery`_.

.. _PyQuery: https://pythonhosted.org/pyquery/index.html
"""
import ast
import re
from setuptools import setup


_version_re = re.compile(r'^__version__\s+=\s+(.*)$', re.MULTILINE)

with open('take/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

with open('README.rst', 'rb') as f:
    long_desc = f.read().decode('utf-8')


setup(
    name='take',
    version=version,
    url='https://github.com/tiffon/take',
    license='MIT',
    author='Joe Farro',
    author_email='joe@jf.io',
    description='A DSL for extracting data from a web page.',
    long_description=long_desc,
    # I hear eggs are bad
    zip_safe=False,
    classifiers=[
        # status of this dist
        'Development Status :: 3 - Alpha',
        # for who
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        # for what
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Utilities',
        # could be a nice companion util with scrapy
        'Framework :: Scrapy',
        # license
        'License :: OSI Approved :: MIT License',
        # env
        'Operating System :: OS Independent',
        # python versions
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='scraping scrapy scraper extraction html',
    packages=['take'],
    install_requires=[
        "cssselect==0.9.1",
        "enum34==1.0.4",
        "lxml==3.4.2",
        "pyquery==1.2.9"
    ],
    include_package_data=True
)
